//*******************************************************************
//*******************************************************************
//***  OCR Engine Custom Proxy to communicate 
//***  with legacy OCR server
//*******************************************************************
//*******************************************************************

//-------------------------------------------------------------------
// Required macros
//-------------------------------------------------------------------
#define UNICODE       // UNICODE
#define _WIN32_DCOM    // DCOM

//-------------------------------------------------------------------
// includes
//-------------------------------------------------------------------
#include <assert.h>
#include <stdio.h>
#include <windows.h>
#include <initguid.h>
#include "..\..\..\basics\OCREngine\idl\ocr.h"

//-------------------------------------------------------------------
// class id of custom proxy
//-------------------------------------------------------------------
// {B4B03F60-DC88-11d1-873B-006008CDD9AE}
static const GUID CLSID_OcrEngineCustomProxy = 
{ 0xb4b03f60, 0xdc88, 0x11d1, { 0x87, 0x3b, 0x0, 0x60, 0x8, 0xcd, 0xd9, 0xae } };


//-------------------------------------------------------------------
// global event telling use wen to shutdown our application
//-------------------------------------------------------------------
HANDLE g_hExitEvent;
wchar_t g_wszModuleName[MAX_PATH];

// Number of living objects
LONG g_lComponentRefCounts = 0; // counts living objects and locks

//-------------------------------------------------------------------
// Component Level Reference count for the lifetime managment of 
// the whole component.
// These specific implementations are used for in-process servers
// There are a pair of functions like these that are used for
// out-of-process servers.
// Return values for debugging purpose only 
//-------------------------------------------------------------------
inline ULONG ComponentAddRef()
{
   return (InterlockedIncrement(&g_lComponentRefCounts));
}

inline ULONG ComponentRelease()  
{
   return (InterlockedDecrement(&g_lComponentRefCounts));
}

//*******************************************************************
//*******************************************************************
//***  UTILITY FUNCTIONS
//*******************************************************************
//*******************************************************************
void DisplayStatus(wchar_t *pwszMsg, HRESULT hr)
{

   if (hr == S_OK) {
      wprintf(TEXT("%s\n"), pwszMsg);
      return;
   }

   if (HRESULT_FACILITY(hr) == FACILITY_WINDOWS) {
      hr = HRESULT_CODE(hr);
   }

   wchar_t *pwszStatus;
   FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
      FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      hr,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      (LPWSTR)&pwszStatus,
      0,
      NULL );

   wprintf(TEXT("%s: %s (ECode: %lx)\n"), pwszMsg, pwszStatus, hr);

   LocalFree(pwszStatus);
}

//-------------------------------------------------------------------
// Given a unicode string, get the number of bytes needed to host it
//-------------------------------------------------------------------
inline long ByteLen(wchar_t *pwsz) 
{
   return (sizeof(wchar_t)*(wcslen(pwsz)+1));
}

//-------------------------------------------------------------------
// Automatic registration; for in-process server, do:
//     regsvr32.exe xxx.dll
//-------------------------------------------------------------------
void RegisterServerToSystem()
{
   wchar_t wszKey[MAX_PATH];
   wchar_t wszValue[MAX_PATH];
   HKEY hKey = 0;

   // CLSID_OcrEngineCustomProxy = {B4B03F60-DC88-11d1-873B-006008CDD9AE}
   // HKEY_CLASSES_ROOT\CLSID\{B4B03F60-DC88-11d1-873B-006008CDD9AE}
   //     @="OcrEngineCustomProxy"
   wcscpy(wszKey, TEXT("CLSID\\{B4B03F60-DC88-11d1-873B-006008CDD9AE}"));
   RegCreateKey(HKEY_CLASSES_ROOT, wszKey, &hKey);
   wcscpy(wszValue, TEXT("OcrEngineCustomProxy"));
   RegSetValueEx(hKey, 0, 0, REG_SZ, (BYTE*)wszValue, ByteLen(wszValue));
   RegCloseKey(hKey);

   // HKEY_CLASSES_ROOT\CLSID\{B4B03F60-DC88-11d1-873B-006008CDD9AE}\InprocServer32
   //     @="E:\DCOM\CODE\whatever.dll"
   //     "ThreadingModel" = "Both"
   wcscpy(wszKey, TEXT("CLSID\\{B4B03F60-DC88-11d1-873B-006008CDD9AE}")
         TEXT("\\InprocServer32"));
   RegCreateKey(HKEY_CLASSES_ROOT, wszKey, &hKey);
   wcscpy(wszValue, g_wszModuleName);
   RegSetValueEx(hKey, 0, 0, REG_SZ, (BYTE*)wszValue, ByteLen(wszValue));
   wcscpy(wszValue, TEXT("Both"));
   RegSetValueEx(hKey, TEXT("ThreadingModel"), 0, REG_SZ, 
              (BYTE*)wszValue, ByteLen(wszValue));
   RegCloseKey(hKey);
}

//-------------------------------------------------------------------
// Automatic registration; for in-process server, do:
//     regsvr32.exe -u xxx.dll
//-------------------------------------------------------------------
void UnregisterServerFromSystem()
{
   long lRc = 0 ;

   lRc=RegDeleteKey(HKEY_CLASSES_ROOT, 
      TEXT("CLSID\\{B4B03F60-DC88-11d1-873B-006008CDD9AE}") 
      TEXT("\\InprocServer32"));
   DisplayStatus(TEXT("Unregistered CLSID's InprocServer"), lRc);

   lRc=RegDeleteKey(HKEY_CLASSES_ROOT, 
      TEXT("CLSID\\{B4B03F60-DC88-11d1-873B-006008CDD9AE}"));
   DisplayStatus(TEXT("Unregistered CLSID"), lRc);
}
 
//*******************************************************************
//*******************************************************************
//***  Class CoOcrEngineCustomProxy
//*******************************************************************
//*******************************************************************
class CoOcrEngineCustomProxy : public IOcr, public IMarshal {
public:
   // constructors/destructors
   CoOcrEngineCustomProxy() : m_lRefCount(0) { ComponentAddRef(); }
   ~CoOcrEngineCustomProxy() { ComponentRelease(); }

   // The factory will call this method to actually 
   // create this object
   static HRESULT CreateObject(LPUNKNOWN pUnkOuter, REFIID riid, 
                        void** ppv);
public: 
   // IUnknown Methods
   STDMETHODIMP QueryInterface(REFIID riid, void **ppv);
   STDMETHODIMP_(ULONG) AddRef(void)
   {
      //------debug messages
      // wchar_t wszMsg[MAX_PATH];
      // wsprintf(wszMsg, TEXT("CoOcrEngineCustomProxy::AddRef (%ld)"), m_lRefCount); 
      // DisplayStatus(wszMsg, S_OK);
      return InterlockedIncrement(&m_lRefCount);
   }
   STDMETHODIMP_(ULONG) Release(void)
   {
      //------debug messages
      // wchar_t wszMsg[MAX_PATH];
      // wsprintf(wszMsg, TEXT("CoOcrEngineCustomProxy::Release (%ld)"), m_lRefCount-1); 
      // DisplayStatus(wszMsg, S_OK);

      long lCount = InterlockedDecrement(&m_lRefCount);
      if (lCount == 0) {
         // DisplayStatus(TEXT("CoOcrEngineCustomProxy::Release - delete object"), S_OK);
         delete this; 
      } 
      return lCount; 
   }

   // IOcr Methods
   STDMETHODIMP OcrImage(long lImageSize, 
                    byte *pbImage,                    
                    wchar_t **pwszOcrText);
   STDMETHODIMP OcrZone(long lImageSize, 
                   byte *pbImage,
                   Zone zone,
                   wchar_t **pwszOcrText);

   // IMarshal Methods
   STDMETHODIMP GetUnmarshalClass(
      REFIID riid,         // IID to be marshaled
      void* pv,            // interface pointer to be marshaled
      DWORD dwDestContext,   // destination - how far away?
      void* pvDestContext,   // reserved
      DWORD mshlflags,      // reason for marshaling
      CLSID* pClsid         // CLSID of custom proxy
   );
   STDMETHODIMP GetMarshalSizeMax(
      REFIID riid,        // IID to be marshaled
      void *pv,          // interface pointer to be marshaled
      DWORD dwDestContext,  // destination - how far away?
      void * pvDestContext, // reserved 
      DWORD mshlflags,     // reason for marshaling
      ULONG * pSize       // returned upper-bound value
   );  
   STDMETHODIMP MarshalInterface(
      IStream *pStm,      // pointer to stream used for marshaling
      REFIID riid,        // IID to be marshaled
      void *pv,          // interface pointer to be marshaled
      DWORD dwDestContext,  // destination - how far away
      void *pvDestContext,  // reserved
      DWORD mshlflags      // reason for marshaling      
   );
   STDMETHODIMP UnmarshalInterface(
      IStream * pStm,  // pointer to the stream to be unmarshaled
      REFIID riid,    // IID to be unmarshaled
      void ** ppv     // returned interface pointer
   );
   STDMETHODIMP ReleaseMarshalData(
     IStream * pStm  // pointer to stream used for unmarshaling
   );
   STDMETHODIMP DisconnectObject(DWORD dwReserved);

private:
   LONG m_lRefCount;
   char m_szLegacyOcrServer[MAX_PATH];
};

//-------------------------------------------------------------------
// CoOcrEngineCustomProxy::CreateObject - static function to create an
// CoOcrEngineCustomProxy object
//-------------------------------------------------------------------
HRESULT CoOcrEngineCustomProxy::CreateObject(LPUNKNOWN pUnkOuter, 
                          REFIID riid, 
                          void** ppv)
{
   *ppv = NULL;

   // CoOcrEngineCustomProxy doesn't support aggregation
   if (pUnkOuter != NULL) { return CLASS_E_NOAGGREGATION; }

   CoOcrEngineCustomProxy * pEngine = new CoOcrEngineCustomProxy;
   if (pEngine == NULL) { return E_OUTOFMEMORY; }

   HRESULT hr = pEngine->QueryInterface(riid, ppv);
   if (FAILED(hr)) { delete pEngine; }

   return hr;
}

//-------------------------------------------------------------------
// CoOcrEngineCustomProxy::QueryInterface
//-------------------------------------------------------------------
STDMETHODIMP
CoOcrEngineCustomProxy::QueryInterface(REFIID riid, void** ppv)
{
   //DisplayStatus(TEXT("CoOcrEngineCustomProxy::QueryInterface"), S_OK);

   if (ppv==NULL) { return E_INVALIDARG; }

   wchar_t *pIID=0;
   StringFromIID(riid, &pIID);
   wprintf(TEXT("CoOcrEngineCustomProxy::QueryInterface[%s]\n"), pIID);
   CoTaskMemFree(pIID);

   if (riid==IID_IUnknown) {
      *ppv= static_cast<IOcr *>(this);
   } else if (riid==IID_IOcr) {
      *ppv= static_cast<IOcr *>(this);
   } else if (riid==IID_IMarshal) {
      *ppv= static_cast<IMarshal *>(this);
   } else {
      *ppv=NULL; return E_NOINTERFACE ;
   }

   reinterpret_cast<IUnknown *>(*ppv)->AddRef();

   return S_OK;
}

//-------------------------------------------------------------------
// IMarshal methods
//-------------------------------------------------------------------
STDMETHODIMP
CoOcrEngineCustomProxy::GetUnmarshalClass(REFIID riid, void* pv,
      DWORD dwDestContext, void* pvDestContext, DWORD mshlflags,
      CLSID* pCustomProxyCLSID)
{
   wprintf(TEXT("CoOcrEngineCustomProxy::GetUnmarshalClass is NOT implemented - apartment model mismatch!\n"));
   return E_NOTIMPL;
}

STDMETHODIMP
CoOcrEngineCustomProxy::GetMarshalSizeMax(REFIID riid, void *pv,
         DWORD dwDestContext, void * pvDestContext, 
         DWORD mshlflags, ULONG * pSize )  
{
   wprintf(TEXT("CoOcrEngineCustomProxy::GetMarshalSizeMax is NOT implemented - apartment model mismatch!\n"));
   return E_NOTIMPL;
}

STDMETHODIMP
CoOcrEngineCustomProxy::MarshalInterface(IStream* pStream, REFIID riid, 
   void *pv, DWORD dwDestContext, void *pvDestContext, DWORD mshlflags)
{
   wprintf(TEXT("CoOcrEngineCustomProxy::MarshalInterface is NOT implemented\n"));
   return E_NOTIMPL;
}

STDMETHODIMP
CoOcrEngineCustomProxy::ReleaseMarshalData(IStream* pStream)
{
   wprintf(TEXT("CoOcrEngineCustomProxy::ReleaseMarshalData is Called\n"));
   return S_OK;
}

STDMETHODIMP
CoOcrEngineCustomProxy::UnmarshalInterface(IStream* pStream, REFIID riid, void** ppv)
{
   wprintf(TEXT("CoOcrEngineCustomProxy::UnmarshalInterface is Called\n"));
   pStream->Read((void *)m_szLegacyOcrServer, MAX_PATH, 0);
   return QueryInterface(riid, ppv);
}

STDMETHODIMP
CoOcrEngineCustomProxy::DisconnectObject(DWORD dwReserved)
{
   wprintf(TEXT("CoOcrEngineCustomProxy::DisconnectObject is NOT implemented\n"));
   return E_NOTIMPL;
}

//-------------------------------------------------------------------
// CoOcrEngineCustomProxy::OcrImage
//-------------------------------------------------------------------
STDMETHODIMP 
CoOcrEngineCustomProxy::OcrImage(long lImageSize, 
                byte *pbImage,                 
                wchar_t **pwszOcrText)
{   
   printf("OcrImage on the legacy server[%s]\n", m_szLegacyOcrServer);
   return S_OK ;
}


//-------------------------------------------------------------------
// CoOcrEngineCustomProxy::OcrZone
//-------------------------------------------------------------------
STDMETHODIMP 
CoOcrEngineCustomProxy::OcrZone(long lImageSize, byte *pbImage, 
               Zone zone, wchar_t **pwszOcrText)
{
   printf("OcrZone on the legacy server[%s]\n", m_szLegacyOcrServer);
   return S_OK ;
}

//*******************************************************************
//*******************************************************************
//***  Class Factory: manufacturing CoOcrEngineCustomProxy objects
//*******************************************************************
//*******************************************************************
class CoOcrEngineCustomProxyFactory : public IClassFactory 
{
public:
   // IUnknown Methods
   STDMETHODIMP QueryInterface (REFIID riid, void** ppv);
   STDMETHODIMP_(ULONG) AddRef(void)
   { return 1; }
   STDMETHODIMP_(ULONG) Release(void) 
   { return 1; }

   // IClassFactory Methods
   STDMETHODIMP CreateInstance(LPUNKNOWN pUnkOuter, 
                        REFIID riid, 
                        void **ppv)
   {
      // call CoOcrEngineCustomProxy's static function to create
      // a CoOcrEngineCustomProxy component object
      return CoOcrEngineCustomProxy::CreateObject(pUnkOuter, riid, ppv);
   }

   STDMETHODIMP LockServer(BOOL fLock)
   {
      if (fLock) {
         ComponentAddRef();
      } else {
         ComponentRelease();
      }
      return S_OK;      
   }   
};

//-------------------------------------------------------------------
//  Singleton factory instance that manufacture CoCorEngine
//  component objects
//-------------------------------------------------------------------
CoOcrEngineCustomProxyFactory g_OcrEngineClassFactory;

//-------------------------------------------------------------------
//  CoOcrEngineCustomProxyFactory::QueryInterface
//-------------------------------------------------------------------
STDMETHODIMP 
CoOcrEngineCustomProxyFactory::QueryInterface(REFIID riid, void** ppv)
{
   if (ppv==NULL) { return E_INVALIDARG; }

   wchar_t *pIID=0;
   StringFromIID(riid, &pIID);
   wprintf(TEXT("CoOcrEngineCustomProxyFactory::QueryInterface[%s]\n"), pIID);
   CoTaskMemFree(pIID);

   if (riid==IID_IUnknown) {
      *ppv= static_cast<IClassFactory *>(this);
   } else if (riid==IID_IClassFactory) {
      *ppv= static_cast<IClassFactory *>(this);
   } else {
      *ppv=NULL; return E_NOINTERFACE ;
   }

   reinterpret_cast<IUnknown *>(*ppv)->AddRef();

   return S_OK;
}

//*******************************************************************
//*******************************************************************
//***  DLL Exports
//*******************************************************************
//*******************************************************************
extern "C" BOOL WINAPI DllMain(HINSTANCE hInstance, 
                        DWORD dwReason, 
                        LPVOID /*lpReserved*/)
{
   if (dwReason == DLL_PROCESS_ATTACH)
   {
      GetModuleFileName(hInstance, g_wszModuleName, MAX_PATH);
   }
   else if (dwReason == DLL_PROCESS_DETACH)
   {
   }
   return TRUE;   // ok
}

//*******************************************************************
//*******************************************************************
//** Used to determine whether the DLL can be unloaded by DCOM
//*******************************************************************
//*******************************************************************
STDAPI DllCanUnloadNow(void)
{ 
   if (g_lComponentRefCounts ==0) {
     return S_OK;
   } else {
     return S_FALSE;
   }
}

//*******************************************************************
//*******************************************************************
//** Returns a class factory to create an object of the requested type
//*******************************************************************
//*******************************************************************
STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv)
{
   *ppv = NULL;

   // Does our component support this particular class ID?
   if (CLSID_OcrEngineCustomProxy!=rclsid) { return E_FAIL; }

   CoOcrEngineCustomProxyFactory * pCF = new CoOcrEngineCustomProxyFactory();
   if (pCF==NULL) { return E_OUTOFMEMORY; }

   HRESULT hr = pCF->QueryInterface(riid, ppv);
   if (FAILED(hr)) {
      delete pCF;
   } else {
      *ppv=pCF;
   }

   return hr;
}

//*******************************************************************
//*******************************************************************
//**  DllRegisterServer - Adds entries to the system registry
//*******************************************************************
//*******************************************************************
STDAPI DllRegisterServer(void)
{
   RegisterServerToSystem();
   return S_OK;
}

//*******************************************************************
//*******************************************************************
//** DllUnregisterServer - Removes entries from the system registry
//*******************************************************************
//*******************************************************************
STDAPI DllUnregisterServer(void)
{
   UnregisterServerFromSystem();
   return S_OK;
}
